import pandas as pd
import numpy as np
from sklearn.metrics import make_scorer, accuracy_score
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
import matplotlib.pyplot as plt
from typing import List, Dict, Tuple
import os
import pyarrow.parquet as pq
from pathlib import Path
import pyarrow.types as pat
import pyarrow as pa
# import dask.dataframe as dd
# from dask_ml.model_selection import KFold
from sklearn.model_selection import cross_val_score

def create_blocks(file: Path, block_size: int, out_dir: Path, step_size: int = 1):
    table = pq.read_table(file)
    rows = []

    if table.num_rows < block_size:
        return []

    print(f'Creating blocks with size {block_size} from {file}')

    for i in range(0, table.num_rows - block_size + 1, step_size):
        block = table.slice(i, block_size)
        last_row = block.slice(block_size - 1, 1)
        row = {
            'TimeStamp': last_row['TimeStamp'][0].as_py(),
            'FrameCounter': last_row['FrameCounter'][0].as_py()
        }

        for column in table.column_names:
            if column not in row.keys():
                column_type = table.schema.field(column).type
                if pat.is_floating(column_type) or pat.is_integer(column_type):
                    column_data = block[column]
                    row[f'{column}_mean'] = pa.compute.mean(column_data).as_py()
                    row[f'{column}_min'] = pa.compute.min(column_data).as_py()
                    row[f'{column}_max'] = pa.compute.max(column_data).as_py()
                else:
                    row[column] = last_row[column][0].as_py()

        rows.append(row)

    df = pd.DataFrame(rows)
    df.set_index(['TimeStamp', 'FrameCounter'], inplace=True)
    df.sort_index(level=['TimeStamp'])

    out_file = out_dir / file.name
    df.to_parquet(out_file)
    return out_file

def split_file_to_maneuvers(file: Path, out_dir: Path) -> List[Path]:
    df = pd.read_parquet(file)
    df['Group'] = (df['Maneuver'] != df['Maneuver'].shift()).cumsum()
    split_dataframes = [group for _, group in df.groupby('Group')]

    out_files: List[Path] = []
    for i in range(len(split_dataframes)):
        df_segment = split_dataframes[i]
        df_segment.drop(columns=['Group'], inplace=True)

        out_file = Path(f'{out_dir / file.stem}_{i}.parquet')
        df_segment.to_parquet(out_file)
        out_files.append(out_file)

    return out_files

def evaluate_block_size(block_size_files: Dict[int, List[Path]]) -> Dict:
    """
    Evaluate different block sizes and return performance metrics
    
    Args:
        data_files: List of tuples containing (parquet_file, json_file) paths
        block_sizes: List of block sizes to evaluate
    """
    results = {}

    for block_size, block_files in block_size_files.items():
        cv_scores = []
        qda = QuadraticDiscriminantAnalysis()
        for file in block_files:
            df = pd.read_parquet(file)
            X_df = df.drop(columns=['Maneuver'])
            y_df = df['Maneuver']
            y = y_df.to_numpy()
            scoring = make_scorer(accuracy_score)
            scores = cross_val_score(qda, X_df, y, cv=5, scoring=scoring)
            cv_scores.append(np.mean(scores))

        results[block_size] = {
            'accuracy': np.mean(cv_scores)
        }
    
    return results

def plot_results(results: Dict):
    """
    Plot the performance metrics for different block sizes
    """
    block_sizes = list(results.keys())
    accuracies = [results[size]['accuracy'] for size in block_sizes]
    
    plt.figure(figsize=(10, 6))
    plt.plot(block_sizes, accuracies, marker='o')
    plt.xlabel('Block Size (time steps)')
    plt.ylabel('Accuracy')
    plt.title('Performance vs Block Size')
    plt.grid(True)
    plt.savefig('out/blocks.png')

if __name__ == '__main__':
    EVALUATION_FILES = [
        'dataset/preprocessing/transformed_data/full_dataset_transformation/removed_unimportant_predictors/92b2d28b-21e4-498c-b6dd-c27a47716a25.parquet',
        'dataset/preprocessing/transformed_data/full_dataset_transformation/removed_unimportant_predictors/39b2c145-c49f-470b-8280-d253fa98153f.parquet',
        'dataset/preprocessing/transformed_data/full_dataset_transformation/removed_unimportant_predictors/ef4852a4-fcfe-429b-b753-d11e2ad08cac.parquet',
    ]
    parquet_files = [Path(f) for f in EVALUATION_FILES]

    block_sizes = [10, 20, 50, 100, 200, 500, 1000]

    block_files: Dict[int, List[Path]] = {}
    for block_size in block_sizes:
        block_files[block_size] = []
        blocks_out_dir = Path(f'out/artifacts/blocks/{block_size}')
        os.makedirs(blocks_out_dir, exist_ok=True)
        for file in parquet_files:
            block_file = create_blocks(file, block_size, step_size=10, out_dir=blocks_out_dir)
            block_files[block_size].append(block_file)
    results = evaluate_block_size(block_files)

    # Plot results
    plot_results(results)

    # Print detailed results
    for block_size, metrics in results.items():
        print(f"\nBlock Size: {block_size}")
        print(f"Accuracy: {metrics['accuracy']:.4f}")
