from pathlib import Path
from typing import List, Dict, Tuple
import pandas as pd

class LazyDataset:
    def __init__(self, files: List[Path]):
        columns = None
        combined_length = 0
        file_start_end: Dict[str, Tuple[int, int]] = {}
        for file in files:
            df = pd.read_parquet(file)
            if columns is not None:            
                columns = df.columns

            df_length = len(df)
            file_start_end[file] = (combined_length, combined_length + df_length - 1)
            combined_length += len(df)

        

        self.n_samples = combined_length
        self.n_features = len(columns)
        self.file_start_end = file_start_end
        self.cached_df = None
    
    def _infer_shape(self):
        return self.n_samples, self.n_features
    
    def __getitem__(self, idx):
        for file, (start, end) in self.file_start_end.items():
            if idx >= start and idx <= end:
                if self.cached_df is not None and self.cached_df[0] == file:
                    df = self.cached_df[1]
                else:
                    df = pd.read_parquet(file)
                    self.cached_df = (file, df)

                return df.iloc[idx - start].to_numpy()
            
        raise OverflowError(f'idx {idx} is out of range, max idx is {self.n_samples}')
    
    def __len__(self):
        return self.n_samples
